@@ -1,5 +1,5 @@
 /*
- *  (C) 2004  Dominik Brodowski <linux@dominikbrodowski.de>
+ *  (C) 2004-2005  Dominik Brodowski <linux@dominikbrodowski.de>
  *
  *  Licensed under the terms of the GNU GPL License version 2.
  */
@@ -9,7 +9,9 @@
 #include <stdio.h>
 #include <errno.h>
 #include <stdlib.h>
+#include <limits.h>
 #include <string.h>
+#include <ctype.h>
 #include <libintl.h>
 #include <locale.h>
 
@@ -23,6 +25,8 @@
 #define gettext_noop(String) String
 #define N_(String) gettext_noop (String)
 
+#define NORM_FREQ_LEN 32
+
 static void print_header(void) {
         printf(PACKAGE_STRING ": cpufreq-set (C) Dominik Brodowski 2004\n");
 	printf(gettext ("Report errors and bugs to %s, please.\n"), PACKAGE_BUGREPORT);
@@ -43,7 +47,9 @@ static void print_help(void) {
 	       "1. Omitting the -c or --cpu argument is equivalent to setting it to zero\n"
 	       "2. The -f FREQ, --freq FREQ parameter cannot be combined with any other parameter\n"
 	       "   except the -c CPU, --cpu CPU parameter\n"
-	       "3. FREQuencies must be passed in kHz =^ MHz * 1000 =^ GHz * 1000000.\n"));
+	       "3. FREQuencies can be passed in Hz, kHz (default), MHz, GHz, or THz\n"
+	       "   by postfixing the value with the wanted unit name, without any space\n"
+	       "   (FREQuency in kHz =^ Hz * 0.001 =^ MHz * 1000 =^ GHz * 1000000).\n"));
 
 }
 
@@ -56,20 +62,107 @@ static struct option set_opts[] = {
 	{ .name="help",		.has_arg=no_argument,		.flag=NULL,	.val='h'},
 };
 
+struct freq_units {
+	char*		str_unit;
+	int		power_of_ten;
+};
+
+const struct freq_units def_units[] = {
+	{"hz", -3},
+	{"khz", 0}, /* default */
+	{"mhz", 3},
+	{"ghz", 6},
+	{"thz", 9},
+	{NULL, 0}
+};
+
 static void print_unknown_arg(void) {
 	print_header();
 	printf(gettext ("invalid or unknown argument\n"));
 	print_help();
 }
 
+static unsigned long string_to_frequency(const char* str)
+{
+	char normalized[NORM_FREQ_LEN];
+	struct freq_units const * unit;	
+	const char* scan;
+	char* end;
+	unsigned long freq;
+	int power = 0, match_count = 0, i, cp, pad;
+
+	while (*str == '0') str++;
+	for (scan = str; isdigit(*scan) || *scan == '.'; scan++) {
+		if (*scan == '.' && match_count == 0)
+			match_count = 1;
+		else if (*scan == '.' && match_count == 1)
+			return 0;
+	}
+
+	if (*scan) {
+		match_count = 0;
+		for (unit = def_units; unit->str_unit; unit++) {
+			for (i = 0;
+			     scan[i] && tolower(scan[i]) == unit->str_unit[i];
+			     ++i);
+			if (scan[i])
+				continue;
+			match_count++;
+			power = unit->power_of_ten;
+		}
+		if (match_count != 1)
+			return 0;
+	}
+
+	/* count the number of digits to be copied */
+	for (cp = 0; isdigit(str[cp]); cp++);
+	if (str[cp] == '.') {
+		while (power > -1 && isdigit(str[cp+1]))
+			cp++, power--;
+	}
+	if (power >= -1)	/* not enough => pad */
+		pad = power + 1;
+	else			/* to much => strip */
+		pad = 0, cp += power + 1;
+	/* check bounds */
+	if (cp <= 0 || cp + pad > NORM_FREQ_LEN - 1)
+		return 0;
+
+	/* copy digits */
+	for (i = 0; i < cp; i++, str++) {
+		if (*str == '.')
+			str++;
+		normalized[i] = *str;
+	}
+	/* and pad */
+	for (; i < cp + pad; i++)
+		normalized[i] = '0';
+
+	/* round up, down ? */
+	match_count = (normalized[i-1] >= '5');
+	/* and drop the decimal part */
+	normalized[i-1] = 0; /* cp > 0 && pad >= 0 ==> i > 0 */
+
+	/* final conversion (and applying rounding) */
+	errno = 0;
+	freq = strtoul(normalized, &end, 10);
+	if (errno)
+		return 0;
+	else {
+		if (match_count && freq != ULONG_MAX)
+			freq++;
+		return freq;
+	}
+}
+
 int main(int argc, char **argv) {
 	extern char *optarg;
 	extern int optind, opterr, optopt;
 	int ret = 0, cont = 1;
 	unsigned int cpu = 0;
-	unsigned long min;
-	unsigned long max;
-	unsigned long freq;
+	unsigned long min = 0;
+	unsigned long max = 0;
+	unsigned long freq = 0;
 	char gov[20];
 	int freq_is_set = 0;
 	int min_is_set = 0;
@@ -107,28 +200,31 @@ int main(int argc, char **argv) {
 			if (min_is_set)
 				double_parm++;
 			min_is_set++;
-			if ((sscanf(optarg, "%lu ", &min)) != 1) {
+			min = string_to_frequency(optarg);
+			if (min == 0) {
 				print_unknown_arg();
 				return -EINVAL;
-                        }
+			}
 			break;
 		case 'u':
 			if (max_is_set)
 				double_parm++;
 			max_is_set++;
-			if ((sscanf(optarg, "%lu ", &max)) != 1) {
+			max = string_to_frequency(optarg);
+			if (max == 0) {
 				print_unknown_arg();
 				return -EINVAL;
-                        }
+			}
 			break;
 		case 'f':
 			if (freq_is_set)
 				double_parm++;
 			freq_is_set++;
-			if ((sscanf(optarg, "%lu ", &freq)) != 1) {
+			freq = string_to_frequency(optarg);
+			if (freq == 0) {
 				print_unknown_arg();
 				return -EINVAL;
-                        }
+			}
 			break;
 		case 'g':
 			if (gov_is_set)
@@ -213,7 +309,7 @@ int main(int argc, char **argv) {
 		       "- Do you have proper administration rights? (super-user?)\n"
 		       "- Is the governor you requested available and modprobed?\n"
 		       "- Trying to set an invalid policy?\n"
-		       "- Trying to set a specific frequency, but userspace governor is not availble,\n"
+		       "- Trying to set a specific frequency, but userspace governor is not available,\n"
 		       "   for example because of hardware which cannot be set to a specific frequency\n"
 		       "   or because the userspace governor isn't loaded?\n"));
 	}
